/*-------------------------------------------------------------------------------
    CRC-16-IBM polynomial (x16 + x15 + x2 + 1 ) with an initialization of 0x0FFFF
 */
var showCRCcalc = false;  // set true to log binary intermediate steps to console

var getBit = function(dataByte, bitPos) {
    return "" + ((dataByte >> bitPos) & 1);
};

var showBits = function(logStr, data, numBits) {
    if (!showCRCcalc) return;
    var msg = "("+numBits + " bits left) : ";
    if (numBits > 0) {
        for (var i = 0; i < data.length; i++) {
            var d = data[i];
            if (numBits >= 8) {
                msg += getBit(d,7) + getBit(d,6) + getBit(d,5) + getBit(d,4) + " ";
                msg += getBit(d,3) + getBit(d,2) + getBit(d,1) + getBit(d,0) + " ";
                numBits -= 8;
            } else {
                var bitsUntilSpace = 4;
                var bitNum = 7;
                while (numBits > 0) {
                    msg += getBit(d,bitNum--);
                    numBits--;
                    bitsUntilSpace--;
                    if (bitsUntilSpace === 0){
                        msg+= " ";
                        bitsUntilSpace = 4;
                    }
                }
            }
        }
    }
    while (logStr.length < 24) {
        logStr += " ";
    }
    console.log(logStr+" "+msg);
};

var flipBitOrder = function(dataByte) {
    var result = ((dataByte & 1) << 7) |
        (((dataByte >> 1) & 1) << 6) |
        (((dataByte >> 2) & 1) << 5) |
        (((dataByte >> 3) & 1) << 4) |
        (((dataByte >> 4) & 1) << 3) |
        (((dataByte >> 5) & 1) << 2) |
        (((dataByte >> 6) & 1) << 1) |
        ((dataByte >> 7) & 1);
    return result;
};

var getFlippedBytes = function(dataBytes) {
    var result = [];
    for (var i=0; i < dataBytes.length; i++) {
        result.push(flipBitOrder(dataBytes[i]));
    }
    return result;
};

var xor17bits = function(dataBytes) {
    var polynomial = 0xC002 << 1 | 1; // 17 bit polynomial
    while (dataBytes.length < 3) {
        dataBytes.push(0);
    }
    var result = dataBytes[0] << 9 | dataBytes[1] << 1;
    if ((dataBytes[2] & 0x80) !== 0) {
        result = result | 1;
    }

    result = result ^ polynomial;
    dataBytes[0] = (result >> 9) & 0x0ff;
    dataBytes[1] = (result >> 1) & 0x0ff;
    dataBytes[2] = dataBytes[2] & 0x07f;
    if (result & 1) {
        dataBytes[2] |= 0x80;
    }
    return dataBytes;
};

var shiftOutLeading0s = function(origData, numBitsRemaining) {
    var numBits = 0;
    var result = [];
    var dataBytes = [...origData];
    var dataByte = 0;
    var done = false;
    while ((dataBytes.length > 0) && (!done)) {
        dataByte = dataBytes[0];
        var bitIndex = 7;
        while ((bitIndex >= 0) && (!done)) {
            if ((dataByte & 0x80) !== 0) {
                done = true;
            } else {
                dataByte = dataByte << 1;
                numBits++;
                bitIndex--;
            }
        }
        if (!done) {
            dataBytes.shift();
        }
    }
    if (done) {
        if ((numBitsRemaining - numBits) < 16) {
            numBits = numBitsRemaining - 16; // keep 16b CRC result
        }
        var numBytesShifted = numBits >> 3;
        while (numBytesShifted > 0) {
            origData.shift();
            numBytesShifted--;
        }
        var numBitsToShift = (numBits % 8);
        var result = [];
        if (numBitsToShift > 0) {
            for (var i = 0; i < origData.length; i++) {
                var nextByte = 0;
                if (i < origData.length - 1){
                    nextByte = origData[i+1];
                }
                var word = (origData[i] << 8) | nextByte;
                var d = (((word << numBitsToShift) >> 8) & 0x0ff);
                result.push(d);
            }
        } else {
            result = origData;
        }
        var numLeft = numBitsRemaining - numBits;
        return {
            dataArray: result,
            numBitsShifted: numBits,
            numBitsRemaining: numLeft
        };
    }
};

var getCRC16IBM = function(dataBytes) {
    var crc = 0x0ffff;
    var CRC16 = 0xC002 << 1 | 1; // 17 bit polynomial
    var buf = getFlippedBytes(dataBytes);
    var numBitsRemaining = buf.length * 8;
    showBits("Flipped", buf, numBitsRemaining);

    buf.push(0);
    buf.push(0);
    numBitsRemaining = buf.length * 8;
    showBits("with 0 CRC bytes", buf, numBitsRemaining);
    buf[0] = buf[0] ^ 0xff;
    buf[1] = buf[1] ^ 0xff;

    showBits("XOR with 0xFFFF", buf, numBitsRemaining);
    var shiftObj = shiftOutLeading0s(buf,numBitsRemaining);
    numBitsRemaining = shiftObj.numBitsRemaining;
    buf = shiftObj.dataArray;
    showBits("shiftOutLeading0s", buf, numBitsRemaining);
    while (numBitsRemaining > 16) {
        buf = xor17bits(buf);
        showBits("xor17", buf, numBitsRemaining);
        var result = shiftOutLeading0s(buf,numBitsRemaining);
        if (result) {
            buf = result.dataArray;
            numBitsRemaining = result.numBitsRemaining;
            showBits("shift out leading 0s", buf, numBitsRemaining);
        } else {
            break;
        }
    }
    // Need to flip the remaining 16b result to normal bit order;
    var crcBuf = getFlippedBytes(buf);
    crc = ((crcBuf[0] & 0xff) << 8) | crcBuf[1] & 0xff;
    return crc;
};

if (showCRCcalc) {
    var testData = [0x80, 0x00, 0x02, 0x0f, 0x0b];
    var test1 = getCRC16IBM(testData);
    console.log("getCRC16IBM: see SLUSC97B section 8.5.1.2.1.5.1 Calculating Frame CRC Value. \nExpected CRC=0xC029.  Was 0x" + test1.toString(16));
}
