var hexArrWithSpaceRegex = /^[a-f 0-9]*$/i;
var hexArrWithoutSpaceRegex = /^[a-f0-9]*$/i;

var positiveIntRegex = /^[0-9]*$/i;

var colorList = [
  'violet',
  'yellow',
  'lavender',
  'cyan',
  'gold',
  'brown',
  'red',
  'pink',
  'magenta',
  'light-brown',
  'green',
  'blue',
  'light-green',
  'light-pink',
  'orange',
  'bright-green'
];

function shallowCopyArr(arr1) {
  var arr2 = [], i;
  for (i = 0; i < arr1.length; i++) {
    arr2[i] = arr1[i];
  }
  return arr2;
}

function shallowCopy(obj1) {
  var keys = Object.keys(obj1);
  var obj2 = {}, i;
  for (i = 0; i < keys.length; i++) {
    obj2[keys[i]] = obj1[keys[i]];
  }
  return obj2;
}

function deepCopy(obj1) {
  var keys = Object.keys(obj1);
  var obj2 = {}, i;
  for (i = 0; i < keys.length; i++) {
      if (typeof obj1[keys[i]] === 'object') {
          obj2[keys[i]] = deepCopy(obj1[keys[i]]);
      } else {
          obj2[keys[i]] = obj1[keys[i]];
      }
  }
  return obj2;
}

function deepCheck(obj1, obj2) {
  var keys = Object.keys(obj1);
  var i, flag = false;;

  for (i = 0; i < keys.length; i++) {
      if (typeof obj1[keys[i]] === 'object') {
          flag = flag || deepCheck(obj1[keys[i]], obj2[keys[i]]);
      } else if (obj1[keys[i]] !== obj2[keys[i]]) {
          flag = true;
      }
  }
  return flag;
}

function prependZero(input, length) {
  while(input.length < length) {
      input = '0' + input;
  }
  return input;
}

function getHexFromInteger(val) {
  return prependZero(val.toString(16).toUpperCase(), 2);
}

function getByteArrFromHexStr(strVal) {
  var arr = [];

  if(strVal.length % 2 === 1) {
    strVal = '0' + strVal;
  }

  while(strVal.length > 0) {
    arr.push(parseInt(strVal.slice(0, 2), 16));
    strVal = strVal.slice(2);
  }

  return arr;
}

function getBitArrFromByte(byte) {
  var arr = byte.toString(2);
  arr = prependZero(arr, 8);
  arr = arr.split('');
  for(var i=0; i<8; i++) {
    arr[i] = parseInt(arr[i]);
  }
  return arr.reverse();
}

function decimalToBoolean(decimal) {
  return decimal === 0 ? false : true;
}

function getCalculatedParamValue(key, value) {
  var multiplier, max15BitValue = 32767, twoPow15 = 32768, max23BitValue = 8388607, twoPow23 = 8388608;

  switch(key) {
    case 'cell':
      multiplier = 0.00019073;
      break;
    case 'busbar':
      multiplier = 0.00003052;
      break;
    case 'bat':
      multiplier = 0.00305;
      break;
    case 'gpio':
      multiplier = 0.00015259;
      break;
    case 'dieTemp':
      multiplier = 0.025;
      break;
    case 'auxGpio':
      multiplier = 0.00015259;
      break;
    case 'diag':
      multiplier = 0.00015259;
      break;
    case 'dac':
      multiplier = 0.00015259 * 1.25 //read value is 0.8 * actual value so we multiply the read value with 1.25 to compensate the loss
      break;
    case 'tsref':
      multiplier = 0.00016954;
      break;
    case 'cs24Bit':
      multiplier = 0.0000000149;
      return value > max23BitValue ?  ((value & max23BitValue) - twoPow23) * multiplier : value * multiplier;
      break;
    default:
      return value;
  }

  if(value > max15BitValue) {
    return ((value & max15BitValue) - twoPow15) * multiplier;
  } else {
    return value * multiplier;
  }
}

function getColor(i) {
  return colorList[i];
}

function getColorCode(name) {
  switch(name) {
    case 'violet':
      return '#5e486e';
    case 'yellow':
      return '#e7e464';
    case 'lavender':
      return '#c8c0e7';
    case 'cyan':
      return '#0bb9cc';
    case 'gold':
      return '#eeb62c';
    case 'brown':
      return '#9e383d';
    case 'red':
      return '#D00';
    case 'pink':
      return '#d5709f';
    case 'magenta':
      return '#7339a3';
    case 'light-brown':
      return '#c4a697';
    case 'green':
      return '#297559';
    case 'blue':
      return '#0c6b90';
    case 'light-green':
      return '#1fc7ab';
    case 'light-pink':
      return '#e9c1d9';
    case 'orange':
      return '#0045fd';
    case 'bright-green':
      return '#98d757';
    default:
      return '#000';
  }
}

function getColorFromCode(colorCode) {
  switch(colorCode) {
    case '#5e486e':
      return 'violet';
    case '#e7e464':
      return 'yellow';
    case '#c8c0e7':
      return 'lavender';
    case '#0bb9cc':
      return 'cyan';
    case '#eeb62c':
      return 'gold';
    case '#9e383d':
      return 'brown';
    case '#D00':
      return 'red';
    case '#d5709f':
      return 'pink';
    case '#7339a3':
      return 'magenta';
    case '#c4a697':
      return 'light-brown';
    case '#297559':
      return 'green';
    case '#0c6b90':
      return 'blue';
    case '#1fc7ab':
      return 'light-green';
    case '#e9c1d9':
      return 'light-pink';
    case '#0045fd':
      return 'orange';
    case '#98d757':
      return 'bright-green';
    default:
      return '#000';
  }
}

function parseDeviceResponse(data) {
  var parsedData = [];

  var dataArr = [], dataArrCount = 0;
  for(i=0; i<data.length; i++) {
      if(dataArr.length === dataArrCount) {
          parsedData.push(dataArr);
          dataArr = [];
          dataArrCount = data[i] + 6 + 1;
      }
      dataArr.push(data[i]);
  }
  parsedData.push(dataArr);
  parsedData.shift();

  return parsedData;
}

function getReadCmdTimeout(noOfReg, noOfDev) {
  return Math.ceil((150 / (134 *  35)) * ((6 + noOfReg) * noOfDev));
  // return Math.ceil(((6 + noOfReg) * noOfDev * 1.1) / 100);
}

function getDateAndTime(dateAndTime) {
  var year = dateAndTime.getFullYear();
  var month = prependZero((dateAndTime.getMonth() + 1).toString(), 2);
  var date = prependZero(dateAndTime.getDate().toString(), 2);

  return year + '-' + month + '-' + date + ' ' + dateAndTime.toLocaleTimeString();
}
function computeCommAddress(devAddr) {
  return (devAddr === "Broadcast" || devAddr === "Stack") ? (devAddr === "Broadcast" ? 0xD0 : 0xB0) : 0x90;
}

function frameCommand(type, devAddr, regAddrbyte1, regAddrbyte2, regValue, cmdArr) {
  console.log(type, devAddr, regAddrbyte1, regAddrbyte2, regValue, cmdArr);
  var writeData = [];
  if(type === 'read') {
    writeData = [0x80, devAddr, regAddrbyte1, regAddrbyte2, regValue];

    var crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
    writeData.push.apply(writeData, crc);
    cmdArr.push(writeData);
    return cmdArr;
  } 
  else if (type === 'write') {
    let commAddr = computeCommAddress(devAddr);

    if(commAddr !== 144)
      writeData = [commAddr, regAddrbyte1, regAddrbyte2, regValue];
    else
      writeData = [commAddr, parseInt(devAddr), regAddrbyte1, regAddrbyte2, regValue];
        
    crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
    writeData.push.apply(writeData, crc);
    cmdArr.push(writeData);
    return cmdArr;
  }
}

function getReadDataLength(cmd, noOfDevices) {
  var readDataLength = 0;

  switch(cmd[0]) {
    case 0x80:
      readDataLength = cmd[4] + 7;
      break;
    case 0xA0:
      readDataLength = (cmd[3] + 7) * (noOfDevices - 1);
      break;
    case 0xC0:
      readDataLength = (cmd[3] + 7) * noOfDevices;
      break;
  }

  return readDataLength;
}