/**
 * Constants
 */
var gc = gc || {};

var numPending = 0;
var maxNumPending = 0;

var UART_9600_bps = 0;
var UART_19200_bps = 1;
var UART_38400_bps = 2;
var UART_57600_bps = 3;
var UART_115200_bps = 4;
var UART_230400_bps = 5;
var UART_300_bps = 6;
var UART_320_bps = 7;
var UART_600_bps = 8;
var UART_1200_bps = 9;
var UART_2400_bps = 10;
var UART_4800_bps = 11;
var UART_460800_bps = 12;
var UART_921600_bps = 13;
var UART_Disable    = 99;

var DEVICE_ADDR = 0x1;


var Cmd_UART_Control = 15; // 0x0F
var Cmd_UART_SetMode = 19; // 0x13

var Cmd_UART_Write = 18; // 0x12
var Cmd_UART_SetMode = 19; // 0x13
var Cmd_UART_Read = 20; // 0x14


var COMMAND_PACKET = 1;
var REPLY_PACKET = 2;
var ERROR_PACKET = 3;
var PAYLOAD_PACKET = 4;
var INTERRUPT_PACKET = 5;

// packet flags
var FLAG_MOREDATA = 1;

var PACKET_IDENTIFIER = 'T'.charCodeAt(0);

var PACKET_ID = 0;
var PACKET_PEC = 1;
var PACKET_PAYLOAD_LEN = 2;
var PACKET_TYPE = 3;
var PACKET_FLAGS = 4;
var PACKET_SEQ_NUM = 5;
var PACKET_STATUS = 6;
var PACKET_COMMAND = 7;
var PACKET_PAYLOAD = 8;

var MAX_PACKET_SIZE = 62; // size, in bytes, of a USB packet
var PACKET_HEADER_SIZE = PACKET_PAYLOAD;
var MAX_PAYLOAD = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;

/**
 * Helper functions
 */

var parseStringProperty = function(message, value, valueMap)
{
    var stringValue = ("" + value).toLowerCase();
    if (valueMap.hasOwnProperty(stringValue))
    {
        return valueMap[stringValue];
    }
    else
    {
        message = TheValue + value + ForEntry + message + InRegDefs + ' is not supported.  Valid entries are';
        var delimiter = ' "';
        var lastOption;
        for(var option in valueMap)
        {
            if (valueMap.hasOwnProperty(option))
            {
                if (lastOption)
                {
                    message = message + delimiter + lastOption;
                    delimiter = '", "';
                }
                lastOption = option;
            }
        }
        throw message + '", or "' + lastOption + '".';
    }
};


var TheValue = 'The value "';
var ForEntry = '", for entry ';
var InRegDefs = ', in the register-defs.json file';
var parseNumberProperty = function (message, value, min, max) {
    var result = +value;
    if (isNaN(result)) {
        throw TheValue + value + ForEntry + message + InRegDefs + ' is not a number.';
    }
    if (min && result < min) {
        throw TheValue + value + ForEntry + message + InRegDefs + ' must be greater than ' + min + '.';
    }
    if (max && result > max) {
        throw TheValue + value + ForEntry + message + InRegDefs + ' must be less than ' + max + '.';
    }
    return result;
};

var getDeviceIndexUart = function(coreIndex, regModelId) {
    coreIndex = coreIndex === undefined ? 0 : coreIndex;
    
    if(coreIndex > -1 && regModelId === 'bq79616' && gc.databind.registry.getModel('bq79600')._codec.getInterfaces()['customuart'].is600) {
        coreIndex = coreIndex + 1;
    }

    return coreIndex;
};

var init = function () {

    var regReadDetail = undefined, regReadCount, regReadPromiseArr;

    /**
     * This code implements an example custom USB2ANY Interface (functionally just a clone of the default I2C interface)
     *
     * @constructor
     * @param settings - The interface settings defined in the system.json
     * @param registerModel - The register model
     * @implements gc.databind.USB2ANY.ICommInterface
     */
    CustomUART_Interface = function (settings, registerModel) {
        this._registerModel = registerModel;
        this.transferPromises = [];
        this.cmdQueue = [];
        this.readyToProcess = true;
        this.noOfDevices = 0;
        this.config = {};
        this.is600 = false;
    };

    /**
     * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
     * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
     * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
     */
    gc.databind.USB2ANY.prototype.CUSTOMUART_Interface = CustomUART_Interface;

    /**
     * Custom interfaces should be extended from the ICommInterface base class
     */
    CustomUART_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

    CustomUART_Interface.prototype._parityMap = { "none": 0, "even": 1, "odd": 2 };
    CustomUART_Interface.prototype._bitDirectionMap = { "lsb": 0, "msb": 1 };

    /**
     * Initialize symbols for device
     *
     * @param settings - The interface settings defined in the system.json
     * @param registerModel - The register model
     */
/*      CustomUART_Interface.prototype.initSymbolsForDevice = function(settings, registerModel) {
         //registerModel.readDeviceAddrsMap('UART', settings);
         var c = gc.databind.PacketCodecFactory.getConstructor(settings.codec);
         if (c && c.prototype && c.prototype.initSymbolsForDevice)
         {
             c.prototype.initSymbolsForDevice(settings, registerModel);
         }
     }; */

    /**
     * Set the communication parameters
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param settings - The interface settings defined in the system.json
     */
    CustomUART_Interface.prototype.control = function (u2a, settings) {
        console.log('UART', 'Control', u2a, settings);

        var baudRate = UART_2400_bps;
        switch (+settings.baudRate) {
            case 19200:
                baudRate = UART_19200_bps;
                break;
            case 38400:
                baudRate = UART_38400_bps;
                break;
            case 57600:
                baudRate = UART_57600_bps;
                break;
            case 115200:
                baudRate = UART_115200_bps;
                break;
            case 230400:
                baudRate = UART_230400_bps;
                break;
            case 300:
                baudRate = UART_300_bps;
                break;
            case 320:
                baudRate = UART_320_bps;
                break;
            case 600:
                baudRate = UART_600_bps;
                break;
            case 1200:
                baudRate = UART_1200_bps;
                break;
            case 2400:
                baudRate = UART_2400_bps;
                break;
            case 4800:
                baudRate = UART_4800_bps;
                break;
            case 460800:
                baudRate = UART_460800_bps;
                break;
            case 921600:
                baudRate = UART_921600_bps;
                break;
        }

        var parity = parseStringProperty('parity', settings.parity, this._parityMap);
        var bitDirection = parseStringProperty('bitDirection', settings.bitDirection, this._bitDirectionMap);
        var characterLength = 8 - parseNumberProperty('characterLength', settings.characterLength || 8, 7, 8);
        var stopBits = parseNumberProperty('stopBits', settings.stopBits || 1, 1, 2) === 1 ? 0 : 1;

        this.config.baudRate = baudRate;
        this.config.parity = parity;
        this.config.bitDirection = bitDirection;
        this.config.characterLength = characterLength;
        this.config.stopBits = stopBits;
    };

    CustomUART_Interface.prototype.configure = function(u2a) {
        this.enable(u2a);

        u2a.onPayloadPacketReceived = this.onPayloadPacketReceived;
        u2a.shouldPauseTransmission = this.shouldPauseTransmission;

        for(var i=0; i<this.transferPromises.length; i++) {
            this.transferPromises[i].reject();
        }
        this.cmdQueue.splice(0, this.cmdQueue.length);
        this.readyToProcess = true;
    };

    CustomUART_Interface.prototype.enable = function(u2a) {
        console.log('UART', 'Config', 'Params', [this.config.baudRate, this.config.parity, this.config.bitDirection, this.config.characterLength, this.config.stopBits]);
        u2a.sendCommandPacket(Cmd_UART_Control,
            [this.config.baudRate, this.config.parity, this.config.bitDirection, this.config.characterLength, this.config.stopBits]
        );

        u2a.sendCommandPacket(Cmd_UART_SetMode, [0]);
    };

    CustomUART_Interface.prototype.disable = function(u2a) {
        u2a.sendCommandPacket(15, [99, 0, 0, 0, 0]);
    };

    CustomUART_Interface.prototype.calculateWriteData = function(writeData) {
        var bytesToRead = 0, bytesToReadHigh, bytesToReadLow, waitTime = 0, waitTimeHigh, waitTimeLow, devAddr;
        var initByte = writeData[0];
        var noOfDevices = this.noOfDevices;

        if(initByte === 0x80) {
            devAddr = writeData[1];
            bytesToRead = writeData[4] + 7;
            // if(writeData[1] === 0x00) {
            //     waitTime = 100 + (bytesToRead * 10);
            // } else {
            //     waitTime = ((devAddr - 1) * 6) + (bytesToRead * 10) + 100;
            // }
        } else if(initByte === 0xA0) {
            bytesToRead = (writeData[3] + 7) * (noOfDevices > 0 ? noOfDevices - 1 : 0);
            // waitTime = ((noOfDevices > 2 ? noOfDevices - 2 : 0) * 6) + (bytesToRead * 10) + 100;
        } else if(initByte === 0xC0) {
            bytesToRead = (writeData[3] + 7) * noOfDevices;
            // waitTime = ((noOfDevices - 1) * 6) + (bytesToRead * 10) + 100;
        }

        bytesToReadHigh = Math.floor(bytesToRead / 256);
        bytesToReadLow = bytesToRead % 256;

        waitTime = 30000; //Wait time in micro seconds
        waitTimeHigh = Math.floor(waitTime / 256);
        waitTimeLow = waitTime % 256;

        writeData.unshift(writeData.length);
        writeData.push(bytesToReadHigh);
        writeData.push(bytesToReadLow);
        writeData.push(waitTimeHigh);
        writeData.push(waitTimeLow);

        return writeData;
    };

    CustomUART_Interface.prototype.shouldPauseTransmission = function(txPacket) {
        return false;
    };

    CustomUART_Interface.prototype.onPayloadPacketReceived = function(target, cmd, rawDataSliced, rawData, payloadLen, payloadStart) {
        // console.log('UART', 'onPayloadPacketReceived', target, cmd, rawData);
        if(cmd === 144 || cmd === 146) {
            var transferPromises = this.getInterfaces()['customuart'].transferPromises;
            // console.log('UART', 'Transfer data in buffer', transferPromises.length);
            if (transferPromises && transferPromises.length > 0) {
                // console.log('UART', 'Transfer promise resolve');
                var transferPromise = transferPromises.shift();
                // console.log('UART', 'Transfer promise shifted', transferPromises.length);
                transferPromise.resolve({rawData: rawData, payloadLen: payloadLen, payloadStart: payloadStart});
            }
        } else if(cmd === 149) {
            console.log('UART', 'onPayloadPacketReceived', target, cmd, rawData, payloadLen);
            var status = rawData.slice(payloadStart, payloadStart + rawData[payloadLen])[0];
            var uartInterface = this.getInterfaces()['customuart'];
            if(status > 0) {
                console.log('UART', 'onPayloadPacketReceived', 'Timeout');
                // uartInterface.disable(this);
                // this.sendCommandPacket(88, [4, 0, 0, 20]);
                // uartInterface.enable(this);
            }
            if(uartInterface.opPromise) {
                console.log('UART', 'onPayloadPacketReceived', 'Resolve');
                uartInterface.opPromise.resolve();
                uartInterface.opPromise = undefined;
            }
        } else {
            gc.console.debug('USB2ANY', 'Ignoring Payload Packet for command #', cmd);
        }
    };

    CustomUART_Interface.prototype.recursivePing = function(u2a, count) {
        var deferred = Q.defer();

        // Clear Rx Buffer
        u2a.sendCommandPacket(147, []);

        // 606
        // var writeData = [0x80, count, 0x01, 0x04, 0x00];

        // 616
        var writeData = [0x80, count, 0x03, 0x06, 0x00];

        var crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
        writeData.push.apply(writeData, crc);
        writeData = this.calculateWriteData(writeData);

        var opPromise = Q.defer();
        this.opPromise = opPromise;

        // setTimeout(function() {
        opPromise.promise.then(function() {
            var buffLenDeferred = Q.defer();
            this.transferPromises.push(buffLenDeferred);
            buffLenDeferred.promise.then(function(data) {
                var rawData = data.rawData;
                var payloadLen = data.payloadLen;
                var payloadStart = data.payloadStart;
                var payload = rawData.slice(payloadStart, payloadStart + rawData[payloadLen]);
                var dataLen = payload[1] * 256 + payload[0];

                if(dataLen > 0) {
                    this.recursivePing(u2a, count + 1).then(function(countData) {
                        deferred.resolve(countData);
                    }.bind(this));
                } else {
                    deferred.resolve(count);
                }
            }.bind(this));
            u2a.sendCommandPacket(146, []);
        // }.bind(this), getReadCmdTimeout(1, 1));
        }.bind(this));

        u2a.sendCommandPacket(149, writeData);

        return deferred.promise;
    }

    CustomUART_Interface.prototype.readFromU2A = function(u2a, readData) {
        var readDeferred = Q.defer();

        var bufferLenDeferred = Q.defer();
        this.transferPromises.push(bufferLenDeferred);
        bufferLenDeferred.promise.then(function(data) {
            // console.log('UART', 'Recursive read', 'Buffer data length', data);
            var rawData = data.rawData;
            var payloadLen = data.payloadLen;
            var payloadStart = data.payloadStart;
            var payload = rawData.slice(payloadStart, payloadStart + rawData[payloadLen]);

            var dataLen = payload[1] * 256 + payload[0];
            var packetSize = Math.ceil(dataLen / 54);

            function packetsTransferResolve(packet) {
                // console.log('UART', 'Recursive read', 'Packet', packet);
                var packetPayloadLen = packet.rawData[packet.payloadLen];
                var packetPayload = packet.rawData.slice(packet.payloadStart, packet.payloadStart + packetPayloadLen);
                readData.push.apply(readData, packetPayload);
                // console.log('UART', 'Recursive read', 'Res data', readData);
            }

            var packetDeferred;
            var packetsTransferPromise = [];
            for(var i=0; i<packetSize; i++) {
                packetDeferred = Q.defer();
                packetsTransferPromise[i] = packetDeferred.promise;
                this.transferPromises.push(packetDeferred);
                packetsTransferPromise[i].then(packetsTransferResolve);
            }
            // console.log('UART', 'Recursive read', 'Packet transfer promises', packetsTransferPromise);
            Promise.all(packetsTransferPromise).then(function() {
                console.log('UART', 'Recursive read', 'Final res', readData);

                readDeferred.resolve(readData);
            }.bind(this));

            if(packetSize > 0) {
                // console.log('UART', 'Recursive read', 'Buffer data transfer', (new Date()).getTime());
                u2a.sendCommandPacket(144, [payload[1], payload[0]]);
            }

        }.bind(this));
        // console.log('UART', 'Recursive read', 'Get buffer size', 'Timestamp', (new Date()).getTime());
        u2a.sendCommandPacket(146, []);

        return readDeferred.promise;
    };

    CustomUART_Interface.prototype.recursiveRead = function(u2a, cmdArr, deferred, readData, calculatedBufferSize) {
        var cmd = cmdArr.shift();

        // console.log('UART', 'Recursive read', 'Calculated read length', getReadDataLength(cmd, this.noOfDevices));

        if(getReadDataLength(cmd, this.noOfDevices) > 4824) {

            this.readFromU2A(u2a, readData).then(function(readData) {

                var tempReadData = [];

                var startReg = cmd[1] * 256 + cmd[2];
                var regCount = cmd[3] + 1;

                var firstStartReg = startReg;
                var firstRegCount = Math.floor(regCount / 2);

                var secondStartReg = startReg + firstRegCount;
                var secondRegCount = regCount - firstRegCount;

                var firstCmd = [cmd[0], cmd[1], cmd[2], firstRegCount - 1];
                var crc = getByteArrFromHexStr(prependZero(getCRC16IBM(firstCmd).toString(16), 4));
                firstCmd.push.apply(firstCmd, crc);
                firstCmd = this.calculateWriteData(firstCmd);

                var secondCmd = [cmd[0], Math.floor(secondStartReg / 256), secondStartReg % 256, secondRegCount - 1];
                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(secondCmd).toString(16), 4));
                secondCmd.push.apply(secondCmd, crc);
                secondCmd = this.calculateWriteData(secondCmd);

                var opPromiseFirst = Q.defer();
                this.opPromise = opPromiseFirst;

                // setTimeout(function() {
                opPromiseFirst.promise.then(function() {
                    this.readFromU2A(u2a, tempReadData).then(function(tempReadData) {

                        var opPromiseSecond = Q.defer();
                        this.opPromise = opPromiseSecond;

                        // setTimeout(function() {
                        opPromiseSecond.promise.then(function() {
                            this.readFromU2A(u2a, tempReadData).then(function(tempReadData) {

                                tempReadData = parseDeviceResponse(tempReadData);

                                for(var i=0; i<tempReadData.length; i++) {
                                    for(var j=i+1; j<tempReadData.length; j++) {
                                        if(tempReadData[i][1] === tempReadData[j][1]) {
                                            tempReadData[i][0] = tempReadData[i][0] + tempReadData[j][0] + 1;
                                            tempReadData[i].splice(tempReadData[i].length - 2, 2);
                                            tempReadData[j].splice(tempReadData[i].length - 2, 2);

                                            for(var k=4; k<tempReadData[j].length; k++) {
                                                tempReadData[i].push(tempReadData[j][k]);
                                            }

                                            crc = getByteArrFromHexStr(prependZero(getCRC16IBM(tempReadData[i]).toString(16), 4));
                                            tempReadData[i].push.apply(tempReadData[i], crc);

                                            tempReadData.splice(j, 1);
                                            j--;
                                        }
                                    }
                                }

                                for(var i=0; i<tempReadData.length; i++) {
                                    readData.push.apply(readData, tempReadData[i]);
                                }

                                // console.log('UART', 'Recursive read', 'Appended data', readData);

                                if(cmdArr.length > 0) {
                                    this.recursiveRead(u2a, cmdArr, deferred, readData, 0);
                                } else {
                                    deferred.resolve(readData);
                                }

                            }.bind(this));
                        }.bind(this));

                        u2a.sendCommandPacket(149, secondCmd);

                    }.bind(this));
                }.bind(this));

                u2a.sendCommandPacket(149, firstCmd);

            }.bind(this));

        } else {
            var noOfRegs = (cmd[0] < 160 ? cmd[4] : cmd[3]) + 1;
            var noOfDevices = cmd[0] < 160 ? 1 : this.noOfDevices;

            calculatedBufferSize = calculatedBufferSize + getReadDataLength(cmd, this.noOfDevices);

            cmd = this.calculateWriteData(cmd);

            if(calculatedBufferSize > 4824) {

                this.readFromU2A(u2a, readData).then(function(readData) {
                    calculatedBufferSize = 0;

                    // console.log('UART', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

                    var opPromise = Q.defer();
                    this.opPromise = opPromise;

                    // setTimeout(function() {
                    opPromise.promise.then(function() {
                        if(cmdArr.length > 0) {
                            this.recursiveRead(u2a, cmdArr, deferred, readData, calculatedBufferSize);
                        } else {
                            this.readFromU2A(u2a, readData).then(function(readData) {
                                deferred.resolve(readData);
                            }.bind(this));
                        }
                    // }.bind(this), getReadCmdTimeout(noOfRegs, noOfDevices));
                    }.bind(this));

                    u2a.sendCommandPacket(149, cmd);

                }.bind(this));

            } else {
                // console.log('UART', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

                var opPromise = Q.defer();
                this.opPromise = opPromise;

                // setTimeout(function() {
                opPromise.promise.then(function() {
                    if(cmdArr.length > 0) {
                        this.recursiveRead(u2a, cmdArr, deferred, readData, calculatedBufferSize);
                    } else {
                        this.readFromU2A(u2a, readData).then(function(readData) {
                            deferred.resolve(readData);
                        }.bind(this));
                    }
                // }.bind(this), getReadCmdTimeout(noOfRegs, noOfDevices));
                }.bind(this));

                u2a.sendCommandPacket(149, cmd);

            }
        }
    };

    // CustomUART_Interface.prototype.recursiveRead = function(u2a, cmdArr, deferred, readData) {
    //     var cmd = cmdArr.shift();
    //     var noOfRegs = (cmd[0] < 160 ? cmd[4] : cmd[3]) + 1;
    //     var noOfDevices = cmd[0] < 160 ? 1 : this.noOfDevices;

    //     cmd = this.calculateWriteData(cmd);
    //     // console.log('SPI', 'Recursive read', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

    //     var opPromise = Q.defer();
    //     this.opPromise = opPromise;

    //     opPromise.promise.then(function() {
    //         if(cmdArr.length > 0) {
    //             this.recursiveRead(u2a, cmdArr, deferred, readData);
    //         } else {
    //             var bufferLenDeferred = Q.defer();
    //             this.transferPromises.push(bufferLenDeferred);
    //             bufferLenDeferred.promise.then(function(data) {
    //                 // console.log('SPI', 'Recursive read', 'Buffer data length', data);
    //                 var rawData = data.rawData;
    //                 var payloadLen = data.payloadLen;
    //                 var payloadStart = data.payloadStart;
    //                 var payload = rawData.slice(payloadStart, payloadStart + rawData[payloadLen]);

    //                 var dataLen = payload[1] * 256 + payload[0];
    //                 var packetSize = Math.ceil(dataLen / 54);

    //                 function packetsTransferResolve(packet) {
    //                     // console.log('SPI', 'Recursive read', 'Packet', packet);
    //                     var packetPayloadLen = packet.rawData[packet.payloadLen];
    //                     var packetPayload = packet.rawData.slice(packet.payloadStart, packet.payloadStart + packetPayloadLen);
    //                     readData.push.apply(readData, packetPayload);
    //                     // console.log('SPI', 'Recursive read', 'Res data', readData);
    //                 }

    //                 var packetDeferred;
    //                 var packetsTransferPromise = [];
    //                 for(var i=0; i<packetSize; i++) {
    //                     packetDeferred = Q.defer();
    //                     packetsTransferPromise[i] = packetDeferred.promise;
    //                     this.transferPromises.push(packetDeferred);
    //                     packetsTransferPromise[i].then(packetsTransferResolve);
    //                 }
    //                 // console.log('SPI', 'Recursive read', 'Packet transfer promises', packetsTransferPromise);
    //                 Promise.all(packetsTransferPromise).then(function() {
    //                     // console.log('SPI', 'Recursive read', 'Final res', readData);

    //                     deferred.resolve(readData);
    //                 }.bind(this));

    //                 if(packetSize > 0) {
    //                     // console.log('SPI', 'Recursive read', 'Buffer data transfer', (new Date()).getTime());
    //                     u2a.sendCommandPacket(144, [payload[1], payload[0]]);
    //                 }

    //             }.bind(this));
    //             // console.log('SPI', 'Recursive read', 'Get buffer size', 'Timestamp', (new Date()).getTime());
    //             u2a.sendCommandPacket(146, []);
    //         }
    //     }.bind(this));

    //     u2a.sendCommandPacket(149, cmd);
    // };

    CustomUART_Interface.prototype.manageReads = function(u2a, cmdArr) {
        var deferred = Q.defer();
        var readData = [];

        // Clear Rx Buffer
        u2a.sendCommandPacket(147, []);

        if(cmdArr.length > 0) {
            this.recursiveRead(u2a, cmdArr, deferred, readData, 0);
        } else {
            deferred.resolve(readData);
        }
        return deferred.promise;
    };

    CustomUART_Interface.prototype.recursiveWrite = function(u2a, cmdArr, deferred) {
        var cmd = cmdArr.shift();
        // console.log('UART', 'Recursive write', 'Timestamp', (new Date()).getTime());
        cmd = this.calculateWriteData(cmd);
        console.log('UART', 'Recursive write', cmd);
        // console.log('UART', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

        var opPromise = Q.defer();
        this.opPromise = opPromise;

        opPromise.promise.then(function() {
            if(cmdArr.length > 0) {
                this.recursiveWrite(u2a, cmdArr, deferred);
            } else {
                deferred.resolve();
            }
        }.bind(this));

        u2a.sendCommandPacket(149, cmd);
    };

    CustomUART_Interface.prototype.manageWrites = function(u2a, cmdArr) {
        var deferred = Q.defer();
        if(cmdArr.length > 0) {
            this.recursiveWrite(u2a, cmdArr, deferred);
        } else {
            deferred.resolve();
        }
        return deferred.promise;
    };

    CustomUART_Interface.prototype.manageReadWrite = function(u2a, cmdInfo) {
        var regPage = document.querySelector('ti-widget-register-page');

        if(cmdInfo) {
            this.cmdQueue.push(cmdInfo);
        }
        // console.log('UART', 'Manage', this.cmdQueue, u2a);
        if(this.readyToProcess && this.cmdQueue.length > 0) {
            var cmd = this.cmdQueue.shift();
            // console.log('UART', 'Manage', 'Cmd', cmd);
            switch(cmd.type) {

                case 'write':
                    console.log('UART', 'Manage', 'Write', cmd);

                    this.readyToProcess = false;
                    this.manageWrites(u2a, [
                        cmd.data
                    ]).then(function() {
                        console.log('UART', 'Manage', 'Write', 'Resolved');
                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));
                    break;

                case 'multipleWrite':
                    console.log('UART', 'Manage', 'Multiple write', cmd);

                    this.readyToProcess = false;
                    this.manageWrites(u2a, cmd.cmdArr).then(function() {
                        if(cmd.promise) {
                            cmd.promise.resolve();
                        }

                        console.log('UART', 'Manage', 'Multiple write', 'Resolved');
                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));
                    break;

                case 'read':
                    console.log('UART', 'Manage', 'Read', cmd);
                    this.readyToProcess = false;

                    if(typeof cmd.data[0] === 'number') {
                        cmd.data = [cmd.data];
                    }

                    this.manageReads(u2a, cmd.data).then(function(resData) {
                        cmd.promise.resolve(resData);

                        console.log('UART', 'Manage', 'Read', 'Resolve');

                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));
                    break;

                case 'poll':
                    console.log('UART', 'Manage', 'Poll', cmd);
                    this.readyToProcess = false;

                    this.manageReads(u2a, [
                        [0xC0, 0x05, 0x36, 0x00, 0x3B, 0x85],  // System Faults
                        [0xC0, 0x05, 0x56, 0x01, 0xD2, 0x45],  // Cell - Balancing Status
                        [0xC0, 0x05, 0xB6, 0x01, 0x9B, 0x85],  // BAT Voltage
                        [0xC0, 0x05, 0x2B, 0x02, 0xB3, 0x14],  // Balancing, Device Status & Fault Summary
                        [0xC0, 0x05, 0x30, 0x02, 0xB9, 0xE4],  // Communication Faults 1, 2, 3
                        [0xC0, 0x05, 0x3C, 0x03, 0x7D, 0x24],  // Cells OVUV
                        [0xC0, 0x05, 0xAE, 0x03, 0x10, 0x44],  // Die Temperature
                        [0xC0, 0x03, 0x18, 0x0F, 0x86, 0x20],  // Cell - Balancing Timer
                        [0xC0, 0x05, 0x8E, 0x0F, 0x09, 0x81]  // GPIO Voltage
                    ]).then(function(resData1) {

                        this.manageReads(u2a, [
                            [0xC0, 0x05, 0x68, 0x21, 0xC3, 0xFD],  // Cell Voltage
                            [0xC0, 0x05, 0xB2, 0x23, 0x19, 0x5C],  // Diagnostics
                            [0xC0, 0x05, 0x8C, 0x01, 0x89, 0x25],  // TSREF
                            [0xC0, 0x05, 0xD6, 0x02, 0xF3, 0x84]   //CS 24bit
                        ]).then(function(resData2) {
                            resData1.push.apply(resData1, resData2);
                            cmd.promise.resolve(resData1);

                            console.log('UART', 'Manage', 'Poll', 'Resolve');
                            this.readyToProcess = true;
                            this.manageReadWrite(u2a);

                        }.bind(this));

                    }.bind(this));
                    break;

                case 'poll-600':
                    console.log('UART', 'Manage', 'Poll - 600', cmd);
                    this.readyToProcess = false;

                    this.manageReads(u2a, [
                        [0xA0, 0x05, 0x36, 0x00, 0x25, 0x85],  // System Faults
                        [0xA0, 0x05, 0x56, 0x01, 0xCC, 0x45],  // Cell - Balancing Status
                        [0xA0, 0x05, 0xB6, 0x01, 0x85, 0x85],  // BAT Voltage
                        [0xA0, 0x05, 0x2B, 0x02, 0xAD, 0x14],  // Balancing, Device Status & Fault Summary
                        [0xA0, 0x05, 0x30, 0x02, 0xA7, 0xE4],  // Communication Faults 1, 2, 3
                        [0xA0, 0x05, 0x3C, 0x03, 0x63, 0x24],  // Cells OVUV
                        [0xA0, 0x05, 0xAE, 0x03, 0x0E, 0x44],  // Die Temperature
                        [0xA0, 0x03, 0x18, 0x0F, 0x98, 0x20],  // Cell - Balancing Timer
                        [0xA0, 0x05, 0x8E, 0x0F, 0x17, 0x81]  // GPIO Voltage
                    ]).then(function(resData1) {

                        this.manageReads(u2a, [
                            [0xA0, 0x05, 0x68, 0x21, 0xDD, 0xFD],  // Cell Voltage
                            [0xA0, 0x05, 0xB2, 0x23, 0x07, 0x5C],  // Diagnostics
                            [0xA0, 0x05, 0x8C, 0x01, 0x97, 0x25],  // TSREF
                            [0xA0, 0x07, 0x80, 0x00, 0xF2, 0x25],  // Debug Comm Stat
                            [0x80, 0x00, 0x21, 0x00, 0x00, 0x75, 0xD4],  // 600 - Fault Summary
                            [0x80, 0x00, 0x23, 0x00, 0x00, 0xD4, 0x14],  // 600 - DEBUG_COMM_STAT
                            [0xA0, 0x05, 0xD6, 0x02, 0xED, 0x84]   //CS 24bit
                        ]).then(function(resData2) {
                            resData1.push.apply(resData1, resData2);
                            cmd.promise.resolve(resData1);

                            console.log('UART', 'Manage', 'Poll - 600', 'Resolve');
                            this.readyToProcess = true;
                            this.manageReadWrite(u2a);

                        }.bind(this));

                    }.bind(this));
                    break;

                case 'autoAddress':
                    console.log('UART', 'Manage', 'Auto address', cmd);
                    this.readyToProcess = false;

                    // 616
                    var writeData, writeDataArr = [], crc;
                    // Broadcast 0x00 for ECC_TEST 0x34C
                    // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                    writeDataArr.push([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                    // Broadcast 0x01 for CONTROL1 0x309
                    // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                    writeDataArr.push([0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                    // Broadcast addr for DIR0_ADDR 0x306
                    for(var i=0; i<36; i++) {
                        writeData = [0xD0, 0x03, 0x06, i];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                        writeData.push.apply(writeData, crc);
                        // u2a.sendCommandPacket(145, writeData);
                        writeDataArr.push(writeData);
                    }

                    this.manageWrites(u2a, writeDataArr).then(function() {

                        this.recursivePing(u2a, 0).then(function(noOfDevices) {
                            console.log('UART', 'No. of devices', noOfDevices);
                            this.noOfDevices = noOfDevices;

                            writeDataArr = []
                            // Broadcast 0x02 for COMM_CTRL 0x308
                            // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            writeDataArr.push([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            if(noOfDevices > 1) {
                                // Single device 0x00 for COMM_CTRL 0x308 at 0x00
                                // u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                                writeDataArr.push([0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                                // Single device 0x03 for COMM_CTRL 0x308 at stack top
                                writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                writeData.push.apply(writeData, crc);
                                // u2a.sendCommandPacket(145, writeData);
                                writeDataArr.push(writeData);
                            } else {
                                // Single device 0x01 for COMM_CTRL 0x308 at 0x00
                                //  u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                                writeDataArr.push([0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                            }

                            this.manageWrites(u2a, writeDataArr).then(function() {

                                this.manageReads(u2a, [
                                    [0xC0, 0x03, 0x4C, 0x00, 0xF8, 0xE4],  // Broadcast read ECC_TEST 0x34C
                                    [0xC0, 0x03, 0x06, 0x00, 0xCF, 0x84]  // Broadcast read DIR0_ADDR 0x306
                                ]).then(function() {
                                    writeDataArr = [];
                                    // Broadcast 0x03 for FAULT_RST2 0x332
                                    // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                                    writeDataArr.push([0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                                    // Broadcast 0x00 for COMM_TIMEOUT_CONF 0x019
                                    // u2a.sendCommandPacket(145, [0xD0, 0x00, 0x13, 0x00, 0x35, 0xD4]);
                                    writeDataArr.push([0xD0, 0x00, 0x19, 0x00, 0x33, 0x74]);

                                    this.manageWrites(u2a, writeDataArr).then(function() {
                                        cmd.promise.resolve(noOfDevices);

                                        console.log('UART', 'Manage', 'Auto address', 'Resolve');
                                        this.readyToProcess = true;
                                        this.manageReadWrite(u2a);
                                    }.bind(this));

                                }.bind(this));

                            }.bind(this));

                        }.bind(this));

                    }.bind(this));
                    break;

                case 'reverse-autoaddress':
                    console.log('UART', 'Manage', 'Reverse auto address', cmd);
                    this.readyToProcess = false;

                    // 616
                    var writeData, writeDataArr = [], crc;
                    // Broadcast 0x00 for ECC_TEST 0x34C
                    // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                    writeDataArr.push([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                    // Sigle device 0x80 for CONTROL1 0x309 at 0x00
                    writeDataArr.push([0x90, 0x00, 0x03, 0x09, 0x80, 0x13, 0xED]);
                    // Reverse broadcast 0x80 for CONTROL1 0x309
                    writeDataArr.push([0xE0, 0x03, 0x09, 0x80, 0xC0, 0x14]);
                    // Broadcast 0x81 for CONTROL1 0x309
                    writeDataArr.push([0xD0, 0x03, 0x09, 0x81, 0x0E, 0xD4]);
                    // Broadcast addr for DIR0_ADDR 0x306
                    for(var i=0; i<36; i++) {
                        writeData = [0xD0, 0x03, 0x07, i];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                        writeData.push.apply(writeData, crc);
                        // u2a.sendCommandPacket(145, writeData);
                        writeDataArr.push(writeData);
                    }

                    this.manageWrites(u2a, writeDataArr).then(function() {

                        this.recursivePing(u2a, 0).then(function(noOfDevices) {
                            console.log('UART', 'No. of devices', noOfDevices);
                            this.noOfDevices = noOfDevices;

                            writeDataArr = []
                            // Broadcast 0x02 for COMM_CTRL 0x308
                            // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            writeDataArr.push([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            if(noOfDevices > 1) {
                                // Single device 0x00 for COMM_CTRL 0x308 at 0x00
                                // u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                                writeDataArr.push([0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                                // Single device 0x03 for COMM_CTRL 0x308 at stack top
                                writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                writeData.push.apply(writeData, crc);
                                // u2a.sendCommandPacket(145, writeData);
                                writeDataArr.push(writeData);
                            } else {
                                // Single device 0x01 for COMM_CTRL 0x308 at 0x00
                                //  u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                                writeDataArr.push([0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                            }

                            this.manageWrites(u2a, writeDataArr).then(function() {

                                this.manageReads(u2a, [
                                    [0xC0, 0x03, 0x4C, 0x00, 0xF8, 0xE4],  // Broadcast read ECC_TEST 0x34C
                                    [0xC0, 0x03, 0x06, 0x00, 0xCF, 0x84]  // Broadcast read DIR0_ADDR 0x306
                                ]).then(function() {
                                    writeDataArr = [];
                                    // Broadcast 0x03 for FAULT_RST2 0x332
                                    // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                                    writeDataArr.push([0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                                    // Broadcast 0x00 for COMM_TIMEOUT_CONF 0x019
                                    // u2a.sendCommandPacket(145, [0xD0, 0x00, 0x13, 0x00, 0x35, 0xD4]);
                                    writeDataArr.push([0xD0, 0x00, 0x19, 0x00, 0x33, 0x74]);

                                    this.manageWrites(u2a, writeDataArr).then(function() {
                                        cmd.promise.resolve(noOfDevices);

                                        console.log('UART', 'Manage', 'Auto address', 'Resolve');
                                        this.readyToProcess = true;
                                        this.manageReadWrite(u2a);
                                    }.bind(this));

                                }.bind(this));

                            }.bind(this));

                        }.bind(this));

                    }.bind(this));
                    break;

                case 'autoAddress-600':
                    console.log('UART', 'Manage', 'Auto address - 600', cmd);
                    this.readyToProcess = false;

                    // 600
                    var writeData, writeDataArr = [], crc;
                    // Single device 0x14 for DEV_CONF1 0x2001 at 0x00
                    // u2a.sendCommandPacket(145, [0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    writeDataArr.push([0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    // Single device 0x20 for CONTROL1 0x0309 at 0x00
                    // u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x09, 0x20, 0x13, 0x95]);
                    // writeDataArr.push([0x90, 0x00, 0x03, 0x09, 0x20, 0x13, 0x95]);

                    this.manageWrites(u2a, writeDataArr).then(function() {

                        // setTimeout(function() {
                            writeDataArr = []
                            // Broadcast 0x00 for ECC_TEST 0x34C
                            // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            writeDataArr.push([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            // Broadcast 0x01 for CONTROL1 0x0309
                            // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                            writeDataArr.push([0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                            // Broadcast addr for DIR0_ADDR 0x306
                            for(var i=0; i<36; i++) {
                                writeData = [0xD0, 0x03, 0x06, i];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                writeData.push.apply(writeData, crc);
                                // u2a.sendCommandPacket(145, writeData);
                                writeDataArr.push(writeData);
                            }

                            this.manageWrites(u2a, writeDataArr).then(function() {

                                this.recursivePing(u2a, 0).then(function(noOfDevices) {
                                    console.log('UART', 'No. of devices', noOfDevices);
                                    this.noOfDevices = noOfDevices;

                                    writeDataArr = []
                                    // Single device write to devices 1, 2, 3… address 0x308 = 0x02
                                    for(var dev = 1; dev < noOfDevices - 1; dev ++) {
                                        writeData = [0x90, dev, 0x03, 0x08, 0x02];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // u2a.sendCommandPacket(145, writeData);
                                        writeDataArr.push(writeData);
                                    }
                                    if(noOfDevices > 1) {
                                        // Single device 0x03 for COMM_CTRL 0x308 at stack top
                                        writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // u2a.sendCommandPacket(145, writeData);
                                        writeDataArr.push(writeData);
                                    }

                                    this.manageWrites(u2a, writeDataArr).then(function() {

                                        this.manageReads(u2a, [
                                            [0xA0, 0x03, 0x4C, 0x00, 0xE6, 0xE4],  // Stack read ECC_TEST 0x306
                                            [0x80, 0x00, 0x20, 0x01, 0x00, 0x25, 0x84],  // Single read read DEV_CONF1 0x2001 at 0x00
                                            [0xA0, 0x03, 0x06, 0x00, 0xD1, 0x84]  // Stack read DIR0_ADDR 0x306
                                        ]).then(function() {
                                            writeDataArr = []
                                            // Stack 0x03 for FAULT_RST2 0x332
                                            // u2a.sendCommandPacket(145, [0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            writeDataArr.push([0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            // Stack 0x00 for COMM_TIMEOUT_CONF 0x019
                                            // u2a.sendCommandPacket(145, [0xB0, 0x00, 0x13, 0x00, 0x2B, 0xD4]);
                                            writeDataArr.push([0xB0, 0x00, 0x19, 0x00, 0x2D, 0x74]);
                                            // Single device 0x00 for COMM_TIMEOUT 0x2005 at 0x00
                                            // u2a.sendCommandPacket(145, [0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);
                                            writeDataArr.push([0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);

                                            this.manageWrites(u2a, writeDataArr).then(function() {
                                                cmd.promise.resolve(noOfDevices);

                                                console.log('UART', 'Manage', 'Auto address- 600', 'Resolve');
                                                this.readyToProcess = true;
                                                this.manageReadWrite(u2a);
                                            }.bind(this));

                                        }.bind(this));

                                    }.bind(this));

                                }.bind(this));

                            }.bind(this));

                        // }.bind(this), 100); // 35 (max no. of 616 stack) * 2.72 (time from shutdown to active + send wake-up)

                    }.bind(this));
                    break;

                case "reverse-autoaddress-600":
                    console.log('UART', 'Manage', 'Reverse auto address - 600', cmd);
                    this.readyToProcess = false;

                    var writeData, writeDataArr = [], crc;
                    //Single device write to 600 CONTROL1[DIR_SEL]=1 
                    // u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x09, 0x80, 0x13, 0xED]);
                    writeDataArr.push([0x90, 0x00, 0x03, 0x09, 0x80, 0x13, 0xED]);
                    //Single device write to 600 DEV_CONF1[SEL_TX_TONE]=0    
                    // u2a.sendCommandPacket(145, [0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    writeDataArr.push([0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    //Single device write to 600 CONTROL1[SEND_WAKE]=1 (wake up stack devices)
                    // u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x09, 0xA0, 0x12, 0x35]);
                    // writeDataArr.push([0x90, 0x00, 0x03, 0x09, 0xA0, 0x12, 0x35]);

                    this.manageWrites(u2a, writeDataArr).then(function() {

                        // setTimeout(function() {
                            writeDataArr = []
                            // Broadcast 0x00 for ECC_TEST 0x34C
                            // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            writeDataArr.push([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            //Broadcast write reverse address 0x309 = 0x80 (change stack devices direction DIR_SEL=1)
                            // u2a.sendCommandPacket(145, [0xE0, 0x03, 0x09, 0x80, 0xC0, 0x14]);
                            writeDataArr.push([0xE0, 0x03, 0x09, 0x80, 0xC0, 0x14]);
                            //Broadcast write address 0x308 = 0x02
                            // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            writeDataArr.push([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            //Broadcast write address 0x309 = 0x81 (enable 616 auto addressing)
                            // u2a.sendCommandPacket(145, [0xD0, 0x03, 0x09, 0x81, 0x0E, 0xD4]);
                            writeDataArr.push([0xD0, 0x03, 0x09, 0x81, 0x0E, 0xD4]);
                            //Broadcast write consecutively address 0x307 = 0, 1, 2, 3… (0 base, rest is stack)
                            for(var i=0; i<66; i++) {
                                writeData = [0xD0, 0x03, 0x07, i];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                writeData.push.apply(writeData, crc);
                                // u2a.sendCommandPacket(145, writeData);
                                writeDataArr.push(writeData);
                            }

                            this.manageWrites(u2a, writeDataArr).then(function() {

                                this.recursivePing(u2a, 0).then(function(noOfDevices) {
                                    console.log('UART', 'No. of devices', noOfDevices);
                                    this.noOfDevices = noOfDevices;

                                    writeDataArr = []
                                    // u2a.sendCommandPacket(145, [0xB0, 0x03, 0x08, 0x02, 0x50, 0xE5]);
                                    for(var dev = 1; dev < noOfDevices; dev ++) {
                                        writeData = [0x90, dev, 0x03, 0x08, 0x02];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // u2a.sendCommandPacket(145, writeData);
                                        writeDataArr.push(writeData);
                                    }
                                    if(noOfDevices > 1) {
                                        // Single device 0x03 for COMM_CTRL 0x308 at stack top
                                        writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // u2a.sendCommandPacket(145, writeData);
                                        writeDataArr.push(writeData);
                                    }

                                    this.manageWrites(u2a, writeDataArr).then(function() {

                                        this.manageReads(u2a, [
                                            [0xA0, 0x03, 0x4C, 0x00, 0xE6, 0xE4],  // Stack read OTC_ECC_TEST  0x34C
                                            [0xA0, 0x03, 0x07, 0x00, 0xD0, 0x14],  // Stack read address 0x307 (read back to verify addresses are correct)
                                            [0x80, 0x00, 0x20, 0x01, 0x00, 0x25, 0x84],  // Single read DEV_CONF1 0x2001 at 0x00
                                        ]).then(function() {
                                            writeDataArr = []
                                            // Stack 0x03 for FAULT_RST2 0x332
                                            // u2a.sendCommandPacket(145, [0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            writeDataArr.push([0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            // Stack 0x00 for COMM_TIMEOUT_CONF 0x019
                                            // u2a.sendCommandPacket(145, [0xB0, 0x00, 0x13, 0x00, 0x2B, 0xD4]);
                                            writeDataArr.push([0xB0, 0x00, 0x19, 0x00, 0x2D, 0x74]);
                                            // Single device 0x00 for COMM_TIMEOUT 0x2005 at 0x00
                                            // u2a.sendCommandPacket(145, [0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);
                                            writeDataArr.push([0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);

                                            this.manageWrites(u2a, writeDataArr).then(function() {
                                                cmd.promise.resolve(noOfDevices);

                                                console.log('UART', 'Manage', 'Reverse auto address - 600', 'Resolve');
                                                this.readyToProcess = true;
                                                this.manageReadWrite(u2a);
                                            }.bind(this));

                                        }.bind(this));

                                    }.bind(this));

                                }.bind(this));

                            }.bind(this));

                        // }.bind(this), 100); // 35 (max no. of 616 stack) * 2.72 (time from shutdown to active + send wake-up)

                    }.bind(this));
                    break;

                case 'debugAutoAddress':
                    console.log('UART', 'Manage', 'Debug auto address', cmd);
                    this.readyToProcess = false;

                    // 616
                    var writeData, crc;
                    var noOfDevices = cmd.noOfDevices;
                    // Broadcast 0x00 for ECC_TEST 0x34C
                    u2a.sendCommandPacket(145, [0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                    // Broadcast 0x01 for CONTROL1 0x309
                    u2a.sendCommandPacket(145, [0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                    // Broadcast addr for DIR0_ADDR 0x306
                    for(var i=0; i<noOfDevices; i++) {
                        writeData = [0xD0, 0x03, 0x06, i];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                        writeData.push.apply(writeData, crc);
                        u2a.sendCommandPacket(145, writeData);
                    }
                    // Broadcast 0x02 for COMM_CTRL 0x308
                    u2a.sendCommandPacket(145, [0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                    if(noOfDevices > 1) {
                        // Single device 0x00 for COMM_CTRL 0x308 at 0x00
                        u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                        // Single device 0x03 for COMM_CTRL 0x308 at stack top
                        writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                        writeData.push.apply(writeData, crc);
                        u2a.sendCommandPacket(145, writeData);
                    } else {
                        // Single device 0x01 for COMM_CTRL 0x308 at 0x00
                        u2a.sendCommandPacket(145, [0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                    }
                    this.manageReads(u2a, [
                        [0xC0, 0x03, 0x4C, 0x00, 0xF8, 0xE4],  // Broadcast read ECC_TEST 0x34C
                        [0xC0, 0x03, 0x06, 0x00, 0xCF, 0x84]  // Broadcast read DIR0_ADDR 0x306
                    ]).then(function(resData) {
                        // console.log('UART', 'Debug auto address', 'Final data', resData);

                        // Broadcast 0x03 for FAULT_RST2 0x332
                        u2a.sendCommandPacket(145, [0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                        // Broadcast 0x00 for COMM_TIMEOUT_CONF 0x019
                        u2a.sendCommandPacket(145, [0xD0, 0x00, 0x19, 0x00, 0x33, 0x74]);

                        cmd.promise.resolve({resData: resData, noOfDevices: cmd.noOfDevices});

                        console.log('UART', 'Manage', 'Debug auto address', 'Resolve');
                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));

                    break;

                case 'readAllBroadcast':
                    console.log('UART', 'Manage', 'Read all broadcast', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 0, readData, crc;

                    function recursive128BroadcastRead() {
                        readData = [];

                        readData.push(0xC0);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('UART', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('UART', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 2048) {
                                recursive128BroadcastRead.bind(this)();
                            } else {
                                console.log('UART', 'Interface', 'Read all', 'Consolidated data', consolidatedData);

                                var regDataArr, regDataRowIndex, regDataColumnIndex, regAddr, rejectFlag;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr;
                                    regDataRowIndex = Math.floor(regAddr/128) * this.noOfDevices;
                                    regDataColumnIndex = regAddr % 128;
                                    regDataArr = [];
                                    rejectFlag = false;
                                    for(var j=0; j<this.noOfDevices; j++) {
                                        // console.log(regAddr, regDataRowIndex + j, regDataColumnIndex + 4);
                                        if(consolidatedData[regDataRowIndex + j]) {
                                            regDataArr.push(consolidatedData[regDataRowIndex + j][regDataColumnIndex + 4]);
                                        } else {
                                            rejectFlag = true;
                                        }
                                    }
                                    if(rejectFlag) {
                                        cmd.promiseArr[i].promise.reject();
                                    } else {
                                        cmd.promiseArr[i].promise.resolve(regDataArr.reverse());
                                        resolveCount++;
                                        console.log('UART', 'Interface', 'Read all - Broadcast', cmd.promiseArr[i].addr, regDataArr);
                                    }
                                }

                                console.log('UART', 'Interface', 'Read all - Broadcast', 'Done', resolveCount);

                                regReadDetail = undefined;

                                regPage.isUpdating = false;
                                console.log('UART', 'Manage', 'Read all broadcast', 'Regmap - Hide progress', regPage);

                                console.log('UART', 'Manage', 'Read all broadcast', 'Resolve');

                                this.readyToProcess = true;
                                this.manageReadWrite(u2a);
                            }
                        }.bind(this));
                    }

                    recursive128BroadcastRead.bind(this)();

                    break;

                case 'readAllStack':
                    console.log('UART', 'Manage', 'Read all stack', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 0, readData, crc;

                    function recursive128StackRead() {
                        readData = [];

                        readData.push(0xA0);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('UART', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('UART', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 2048) {
                                recursive128StackRead.bind(this)();
                            } else {
                                console.log('UART', 'Interface', 'Read all - Stack', 'Consolidated data', consolidatedData);

                                var regDataArr, regDataRowIndex, regDataColumnIndex, regAddr, rejectFlag;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr;
                                    regDataRowIndex = Math.floor(regAddr/128) * (this.noOfDevices - 1);
                                    regDataColumnIndex = regAddr % 128;
                                    regDataArr = [];
                                    rejectFlag = false;
                                    for(var j=0; j<this.noOfDevices - 1; j++) {
                                        // console.log(regAddr, regDataRowIndex + j, regDataColumnIndex + 4);
                                        if(consolidatedData[regDataRowIndex + j]) {
                                            regDataArr.push(consolidatedData[regDataRowIndex + j][regDataColumnIndex + 4]);
                                        } else {
                                            rejectFlag = true;
                                        }
                                    }
                                    if(rejectFlag) {
                                        cmd.promiseArr[i].promise.reject();
                                    } else {
                                        cmd.promiseArr[i].promise.resolve(regDataArr.reverse());
                                        resolveCount++;
                                        console.log('UART', 'Interface', 'Read all - Stack', cmd.promiseArr[i].addr, regDataArr);
                                    }
                                }

                                console.log('UART', 'Interface', 'Read all - Stack', 'Done', resolveCount);

                                regReadDetail = undefined;

                                regPage.isUpdating = false;
                                console.log('UART', 'Manage', 'Read all stack', 'Regmap - Hide progress', regPage);

                                console.log('UART', 'Manage', 'Read all stack', 'Resolve');

                                this.readyToProcess = true;
                                this.manageReadWrite(u2a);
                            }
                        }.bind(this));
                    }

                    recursive128StackRead.bind(this)();

                    break;

                case 'readAllSingle':
                    console.log('UART', 'Manage', 'Read all single', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 0, readData, crc;

                    function recursive128SingleRead() {
                        readData = [];

                        readData.push(0x80);
                        readData.push(cmd.device);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('UART', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('UART', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 2048) {
                                recursive128SingleRead.bind(this)();
                            } else {
                                console.log('UART', 'Interface', 'Read all - Single', 'Consolidated data', consolidatedData);

                                var regDataRowIndex, regDataColumnIndex, regAddr;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr;
                                    regDataRowIndex = Math.floor(regAddr/128);
                                    regDataColumnIndex = regAddr % 128;
                                    // console.log(regAddr, regDataRowIndex, regDataColumnIndex + 4);
                                    if(consolidatedData[regDataRowIndex]) {
                                        cmd.promiseArr[i].promise.resolve(consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                        resolveCount++;
                                        console.log('UART', 'Interface', 'Read all - Single', cmd.promiseArr[i].addr, consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                    } else {
                                        cmd.promiseArr[i].promise.reject();
                                    }
                                }

                                console.log('UART', 'Interface', 'Read all - Single', 'Done', resolveCount);

                                regReadDetail = undefined;

                                regPage.isUpdating = false;
                                console.log('UART', 'Manage', 'Read all single', 'Regmap - Hide progress', regPage);

                                console.log('UART', 'Manage', 'Read all single', 'Resolve');

                                this.readyToProcess = true;
                                this.manageReadWrite(u2a);
                            }
                        }.bind(this));
                    }

                    recursive128SingleRead.bind(this)();

                    break;

                case 'readAllSingle600':
                    console.log('UART', 'Manage', 'Read all single - 600', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 8192, readData, crc;

                    function recursive128SingleRead600() {
                        readData = [];

                        readData.push(0x80);
                        readData.push(cmd.device);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('UART', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('UART', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 9216) {
                                recursive128SingleRead600.bind(this)();
                            } else {
                                console.log('UART', 'Interface', 'Read all - 600', 'Consolidated data', consolidatedData);

                                var regDataRowIndex, regDataColumnIndex, regAddr;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr - 8192;
                                    regDataRowIndex = Math.floor(regAddr/128);
                                    regDataColumnIndex = regAddr % 128;
                                    // console.log(regAddr, regDataRowIndex, regDataColumnIndex + 4);
                                    if(consolidatedData[regDataRowIndex]) {
                                        cmd.promiseArr[i].promise.resolve(consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                        resolveCount++;
                                        console.log('UART', 'Interface', 'Read all - 600', cmd.promiseArr[i].addr, consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                    } else {
                                        cmd.promiseArr[i].promise.reject();
                                    }
                                }

                                readData = [0x80, cmd.device, 0x03, 0x06, 0x04];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                                readData.push.apply(readData, crc);

                                this.manageReads(u2a, [readData]).then(function(resData2) {
                                    // var data = parseDeviceResponse(resData2)[0];
                                    var data = resData2;
                                    console.log('UART', 'Interface', 'Read all - 600', 'Consolidated data 2', data);
                                    var addr, val;

                                    for(var i=0; i<cmd.promiseArr.length; i++) {
                                        addr = cmd.promiseArr[i].addr;

                                        if(addr > 0x305 && addr < 0x30B) {
                                            val = data[addr - 0x306 + 4];
                                            if(val) {
                                                cmd.promiseArr[i].promise.resolve(val);
                                            } else {
                                                cmd.promiseArr[i].promise.reject();
                                            }
                                            resolveCount++;
                                            console.log('UART', 'Interface', 'Read all - 600', cmd.promiseArr[i].addr, val);
                                        }
                                    }

                                    console.log('UART', 'Interface', 'Read all - Single - 600', 'Done', resolveCount);

                                    regReadDetail = undefined;

                                    regPage.isUpdating = false;
                                    console.log('UART', 'Manage', 'Read all single - 600', 'Regmap - Hide progress', regPage);

                                    console.log('UART', 'Manage', 'Read all single - 600', 'Resolve');

                                    this.readyToProcess = true;
                                    this.manageReadWrite(u2a);
                                }.bind(this));

                            }
                        }.bind(this));
                    }

                    recursive128SingleRead600.bind(this)();

                    break;

                case 'cb-remaining-time':
                  var remainingTimeArr = [];
                  this.readyToProcess = false;
                    function readRemainingTime(cellNo) {
                      writeDataArr = [];
                      writeData = [0x90, cmd.device, 0x03, 0x30, (cellNo*2)+1];
                      crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                      writeData.push.apply(writeData, crc);
                      writeDataArr.push(writeData);
                      this.manageWrites(u2a,writeDataArr).then(function () {

                        readData = [0x80, cmd.device, 0x05, 0x58, 0x00];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        this.manageReads(u2a, [readData]).then(function (data) {
                          remainingTimeArr.push(data[4]);
                          if(cellNo < 15) {
                            readRemainingTime.bind(this)(cellNo+1);
                          } else {
                            cmd.promise.resolve(remainingTimeArr);
                            remainingTimeArr = [];  
                            this.readyToProcess = true;
                            this.manageReadWrite(u2a);

                          }
                        }.bind(this));
                      }.bind(this));
                    }

                    readRemainingTime.bind(this)(0);
                    
                    break;
            }
        }
    }

    /**
     * Construct and send a command packet for reading from a register
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param info - The register map info for the register
     * @param regModel - the register model that is requesting the read
     * @param coreIndex - the core to read (undefined if not a deviceArray)
     * @return - A promise that is resolved once the data is received
     */
    CustomUART_Interface.prototype.read = function(u2a, info, regModel, coreIndex) {
        console.log('UART', 'Read', u2a, info, regModel, coreIndex, getDeviceIndexUart(coreIndex, regModel._id), regReadDetail);
        var regAddr = info.addr;
        var readData = [], crc;

        var deviceIndex = getDeviceIndexUart(coreIndex, regModel._id);

        var regPage = document.querySelector('ti-widget-register-page');
 
        if(regReadDetail) {
        // if(regReadDetail && ((regReadDetail.regModelId === regModel._id) && ((coreIndex === undefined) || (regReadDetail.coreIndex === coreIndex)))) {

            if(deviceIndex < 0) {
                var deferred = Q.defer();

                regReadPromiseArr.push({
                    promise: deferred,
                    addr: info.addr
                });

                regReadCount++;

                if(regReadCount === regReadDetail.numRegs) {
                    this.manageReadWrite(u2a, {
                        type: this.is600 === true ? (regModel._id === 'bq79600' ? 'readAllBroadcast600' : 'readAllStack') : 'readAllBroadcast',
                        device: deviceIndex,
                        count: regReadCount,
                        promiseArr: regReadPromiseArr
                    });
                }

                return deferred.promise;
            } else {
                var deferred = Q.defer();

                regReadPromiseArr.push({
                    promise: deferred,
                    addr: info.addr
                });

                regReadCount++;

                if(regReadCount === regReadDetail.numRegs) {
                    this.manageReadWrite(u2a, {
                        type: regModel._id === 'bq79600' ? 'readAllSingle600' : 'readAllSingle',
                        device: deviceIndex,
                        count: regReadCount,
                        promiseArr: regReadPromiseArr
                    });
                }

                return deferred.promise;
            }

        } else {

            regPage.isUpdating = true;
            console.log('UART', 'Read', 'Regmap - Show progress', regPage);

            if(deviceIndex < 0) {
                readData.push(0xC0);    // Broadcast read

                regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
                readData.push.apply(readData, regAddr);    // Register address
                readData.push(0x00);    // Read 1 byte of data
                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                readData.push.apply(readData, crc);

                var deferred = Q.defer();

                var readPromise = Q.defer();
                readPromise.promise.then(function(data) {
                    var readData = [];
                    data = parseDeviceResponse(data);
                    for(var i=0; i<data.length; i++) {
                        readData.push(data[i][4]);
                    }
                    readData = readData.reverse();
                    deferred.resolve(readData);
                    regPage.isUpdating = false;
                    console.log('UART', 'Read', 'Broadcast read', 'Regmap - Hide progress', regPage);
                })
                this.manageReadWrite(u2a, {type: 'read', data: readData, promise: readPromise});

                return deferred.promise;
            } else {
                readData.push(0x80);    // Single device read intialization byte
                readData.push(deviceIndex);    // Device address
                regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
                readData.push.apply(readData, regAddr);    // Register address
                readData.push(0x00);    // Read 1 byte of data
                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                readData.push.apply(readData, crc);    // CRC bytes

                var deferred = Q.defer();

                var readPromise = Q.defer();
                readPromise.promise.then(function(data) {
                    deferred.resolve(data[4]);
                    regPage.isUpdating = false;
                    console.log('UART', 'Read', 'Single read', 'Regmap - Hide progress', regPage);
                })
                this.manageReadWrite(u2a, {type: 'read', data: readData, promise: readPromise});

                return deferred.promise;
            }

        }
    };

    /**
     * Construct and send a command packet for writing to a register
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param info - The register map info for the register
     * @param value - The value to write to the register
     * @param regModel - the register model that is requesting the write
     * @param coreIndex - the core to write (undefined if not a deviceArray)* 
     * @return - A promise that resolves immediately
     */
    CustomUART_Interface.prototype.write = function(u2a, info, value, regModel, coreIndex) {
        console.log('UART', 'Write', u2a, info, value, regModel, coreIndex);
        var writeData = [];
        var regAddr, crc;

        var deviceIndex = getDeviceIndexUart(coreIndex, regModel._id);

        if(deviceIndex > -1) {
            writeData.push(0x90);    // Single device write intialization byte
            writeData.push(deviceIndex);    // Device address
            regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
            writeData.push.apply(writeData, regAddr);    // Register address
            writeData.push(value);    // Data byte
            crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
            writeData.push.apply(writeData, crc);    // CRC bytes

            console.log('UART', 'Registermap', 'Write', writeData);
            this.manageReadWrite(u2a, {type: 'write', data: writeData});
        }

        return Q(true);
    };

    var readRegsHdlr = function(eventType, event) {
        console.log('UART', 'Reg read event', eventType, event, event.detail);

        var detailObj = event.detail;
        if (detailObj.type === undefined) {
            detailObj.type = eventType;
        }

        var modelId = undefined;
        if (detailObj.regModelId) {
            modelId = detailObj.regModelId;
        }

        var deviceIndex = getDeviceIndexUart(detailObj.coreIndex, modelId);

        console.log("bc: read_regs: numRegs="+detailObj.numRegs+", regModel id="+modelId+", coreIndex="+detailObj.coreIndex+", type="+detailObj.type);
        if(detailObj.type === 'init' || detailObj.type === 'button' || detailObj.type === 'refresh') {
            // setTimeout(function() {
                // var modelId = "undefined";

                var regPage = document.querySelector('ti-widget-register-page');
                regPage.isUpdating = true;
                console.log('UART', 'Reg read event', 'Regmap - Show progress', regPage);

                if (regReadDetail && regReadDetail.modelId) {
                    modelId = regReadDetail.modelId;
                }
                console.log("regReadDetail.regModelId changed from "+modelId +" to "+detailObj.regModelId);
                regReadDetail = detailObj;
                regReadDetail.device = deviceIndex;
                regReadCount = 0;
                regReadPromiseArr = [];
            // },0);
        }
    }

    var refreshPage = function() {
        var regPage = document.querySelector('ti-widget-register-page');
        regPage.readAllRegistersClickHdlr();
        templateObj.timerHdlr = window.setTimeout(refreshPage, templateObj.refreshInMs);
    };

    var templateObj = document.querySelector('#template_obj');
    templateObj.timerHdlr = null;
    templateObj.refreshInMs = 100;

    templateObj.addEventListener('dom-change', function(event) {
        if(document.querySelector('ti-bq79616-global')) {
            document.addEventListener("refresh-regs", readRegsHdlr.bind(templateObj,"refresh"));
            document.addEventListener("init-regs", readRegsHdlr.bind(templateObj,"init"));
            document.addEventListener("read_regs", readRegsHdlr.bind(templateObj,"statusOrButton"));
            document.addEventListener("read_reg", function(event) {
                regReadDetail = undefined;
            });
            document.addEventListener("auto_read_config_changed", function(event) {
               if(templateObj.timerHdlr) {
                   window.clearTimeout(templateObj.timerHdlr);
                   templateObj.timerHdlr = null;
               }
               var autoReadObj = event.detail.autoRead;
               if (autoReadObj.value > -1) {
                   templateObj.refreshInMs = +autoReadObj.value;
                   templateObj.timerHdlr = window.setTimeout(refreshPage, templateObj.refreshInMs);
               }
            });
        }
    });

};


// Wait until USB2ANY global var is ready before defining interface
if (gc.databind.USB2ANY) {
    init();
} else {
    document.addEventListener('DOMContentLoaded', init);
}