/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};
var saveAsFilePath; // undefined initially

/*
 *  Boilerplate code for creating computed data bindings
 */
document.addEventListener('gc-databind-ready', function() {
    /*
     *   Add custom computed value databindings here, using the following method:
     *
     *   function gc.databind.registry.bind(targetBinding, modelBinding, [getter], [setter]);
     *
     *   param targetBinding - single binding string or expression, or array of binding strings for multi-way binding.
     *   param getter - (optional) - custom getter function for computing the targetBinding value(s) based on modelBinding value(s).
     *   param setter - (optional) - custom setter function for computing the modelBinding value(s) based on targetBinding value(s).
     */

    // For example, a simple computed values based on simple expression
    // gc.databind.registry.bind('widget.id.propertyName', "targetVariable == 1 ? 'binding is one' : 'binding is not one'");

    // Or a custom two-way binding with custome getter and setter functions.  (setter is optional)  (getter only indicates one-way binding)
    // gc.databind.registry.bind('widget.id.propertyName', "targetVariable", function(value) { return value*5/9 + 32; }, function(value) { (return value-32)*9/5; });

    // Event 1 to n bindings
    /*
    gc.databind.registry.bind('widget.date.value',
        // dependant bindings needed in order to compute the date, in name/value pairs.
        {
            weekday: 'widget.dayOfWeek.selectedText',
            day: 'widget.dayOfMonth.value',
            month: 'widget.month.selectedText',
            year: 'widget.year.value'
        },
        // getter for date computation
        function(values)
        {
            // compute and return the string value to bind to the widget with id 'date'
            return values.weekday + ', ' + values.month + ' ' + values.day + ', ' + values.year;
        }
    );
	*/
});

/*
 *  Boilerplate code for creating custom actions
 */
document.addEventListener('gc-nav-ready', function() {
    /*
     *   Add custom actions for menu items using the following api:
     *
     *   function gc.nav.registryAction(id, runable, [isAvailable], [isVisible]);
     *
     *   param id - uniquely identifies the action, and should correspond to the action property of the menuaction widget.
     *   param runable - function that performs the custom action.
     *   param isAvailable - (optional) - function called when the menu action is about to appear.  Return false to disable the action, or true to enable it.
     *   param isVisible - (optional) - function called when the menu action is about to appear.  Return false to hide the action, or true to make it visible.
     */

    // For example,
    // gc.nav.registerAction('myCustomCloseAction', function() { window.close(); }, function() { return true; }, function() { return true; });

    // Alternatively, to programmatically disable a menu action at any time use:
    // gc.nav.disableAction('myCustomCloseAction);    then enable it again using:  gc.nav.enableAction('myCustomCloseAction');
  
    gc.nav.registerAction('show_introduction', function() {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_vtabcontainer.selectedIndex = 0;
            templateObj.$.ti_widget_menuitem.hideFullMenu();
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });
    gc.nav.registerAction('show_registers', function() {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_vtabcontainer.selectedIndex = 1;
            templateObj.$.ti_widget_menuitem.hideFullMenu();
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });
    gc.nav.registerAction('show_motor_control', function() {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_vtabcontainer.selectedIndex = 2;
            templateObj.$.ti_widget_menuitem.hideFullMenu();
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });

    gc.nav.registerAction('load_registers', function() {
        if ((templateObj) && (templateObj.$)) {
            gc.File.browseAndLoad(null, {
                bin: false
            }, function(contents, properties) {
                var regData = null;
                try {
                    regData = JSON.parse(contents);
                    if (regData) {
                        templateObj.setRegisterData(regData);
                        templateObj.$.ti_widget_eventlog_view.log("info", "Load Registers: loaded from :" + properties.localPath);
                    }
                } catch (ex) {
                    templateObj.logMsgAndNotify("error", "Load Registers: Could not parse file:" + ex, false, true);
                }

            });
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });

    var saveText = function(fname, text, mime) {
        var blob = new Blob([text], {
            type: mime
        });
        saveAs(blob, fname);
        return false;
    };

    var saveRegsAs = function(calledBy, regDataJson) {
        gc.File.browseAndSave(regDataJson, 'regs.json', '.json', function(result, errorInfo) {
            if (errorInfo) {
                templateObj.logMsgAndNotify("error", calledBy + " " + errorInfo, false, true);
            } else if ((result) && (result.localPath)) {
                saveAsFilePath = result.localPath;
                templateObj.$.ti_widget_eventlog_view.log("info", "Registers saved to " + result.localPath);
            }
        });
    };

    gc.nav.registerAction('save_registers', function() {
        if ((templateObj) && (templateObj.$)) {
            var regData = templateObj.getRegisterData();
            var regDataJson = JSON.stringify(regData);
            if (gc.desktop.isDesktop()) {
                if (!saveAsFilePath) {
                    saveRegsAs("Save Registers", regDataJson);
                } else {
                    gc.File.save(regDataJson, {"localPath": saveAsFilePath}, null, function(errorInfo) {
                        if (errorInfo) {
                            templateObj.logMsgAndNotify("error", "Save Registers As: " + errorInfo, false, true);
                        } else {
                            templateObj.$.ti_widget_eventlog_view.log("info", "Registers saved to " + saveAsFilePath);
                        }
                    });
                }

            } else {
                saveText("regs.json", regDataJson, 'text/plain;charset=utf-8');
                templateObj.$.ti_widget_eventlog_view.log("info", "Save Registers: regs.json file downloaded");
            }
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });

    gc.nav.registerAction('save_registers_as', function() {
        if ((templateObj) && (templateObj.$)) {
            var regData = templateObj.getRegisterData();
            var regDataJson = JSON.stringify(regData);
            if (gc.desktop.isDesktop()) {
                saveRegsAs("Save Registers As", regDataJson);
            } else {
                saveText("regs.json", regDataJson, 'text/plain;charset=utf-8');
                templateObj.$.ti_widget_eventlog_view.log("info", "Save Registers As: regs.json file downloaded");
            }
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });

    gc.nav.registerAction('open_log_pane', function() {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_eventlog_view.openView();
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });

    gc.nav.registerAction('open_scripting_dialog', function() {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.scripting_dialog.removeAttribute('hidden');
            templateObj.$.scripting_dialog.open();
        }
    }, function() {
        return true;
    }, function() {
        return true;
    });

});

/*
 *  Boilerplate code for working with components in the application gist
 */


var initComplete = false;
var templateObj;
var registerModel;
var bmsGlobalObj;


// Wait for DOMContentLoaded event before trying to access the application template
var init = function() {
    templateObj = document.querySelector('#template_obj');

    // Wait for the template to fire a dom-change event to indicate that it has been 'stamped'
    // before trying to access components in the application.
    templateObj.addEventListener('dom-change', function() {
        if (initComplete) return;
        this.async(function() {
            initComplete = true;
            console.log("Application template has been stamped.");
            bmsGlobalObj = document.querySelector('ti-bq79616-global');
            templateObj.$.ti_widget_toast.hideToast();
            templateObj.$.ti_widget_eventlog_view.log("info", "Application started.");
            registerModel = Polymer.dom(document).querySelector('ti-model-register');
            if (registerModel) {
                registerModel.addEventListener("usbHidModelReceivingData", function (event) {
                    templateObj.logMsgAndNotify("debug", "Receiving data from target.", false, false);
                });
            } else {
                console.log("no ti-model-register component found in this application!")
            }

            // Now that the template has been stamped, you can use 'automatic node finding' $ syntax to access widgets.
            // e.g. to access a widget with an id of 'widget_id' you can use templateObj.$.widgetId

            if (templateObj.$.ti_widget_tabcontainer_guide) {
                templateObj.$.ti_widget_tabcontainer_guide.selectedIndex = 0; // always start with the home page

                templateObj.$.ti_widget_label_expert.addEventListener("click", function(event) {
                    templateObj.$.ti_widget_vtabcontainer.selectedIndex = 2;
                });

                templateObj.$.ti_widget_label_beginner.addEventListener("click", function(event) {
                    templateObj.$.ti_widget_tabcontainer_guide.selectedIndex = 1;
                });

                templateObj.skipPage = function() {
                    templateObj.$.ti_widget_tabcontainer_guide.selectedIndex = 1 + templateObj.$.ti_widget_tabcontainer_guide.selectedIndex;
                };

                templateObj.$.ti_widget_label_skip_page.addEventListener("click", templateObj.skipPage);

		        // when the user changes the vtab, reset the guide back to its first page
                templateObj.$.ti_widget_vtabcontainer.addEventListener("selected_tab_changed", function(event) {
                    if (templateObj.$.ti_widget_vtabcontainer === 1) {
                        // reset the guide back to the first tab
                        templateObj.$.ti_widget_tabcontainer_guide.selectedIndex = 0;
                    }
                });
            }
            // open the eventlog viewer when the user clicks on the book icon in the status bar:
            templateObj.$.ti_widget_eventlog_view.closeView(); // ensure that view states are in sync with application
            templateObj.$.ti_widget_statusbar.addEventListener("appiconclicked", function(event) {
                templateObj.$.ti_widget_eventlog_view.toggleView();
            });

            var aboutBox = document.querySelector('ti-widget-aboutbox');
            if (aboutBox) {
                aboutBox.addEventListener('aboutbox_opening', function(event) {
                    aboutBox.softwareManifestLink = "docs/bq79616 EVM GUI_1.0_manifest.html";
                    aboutBox.applicationVersion = "2.0.0";
                    aboutBox.licenseLink = "docs/license.pdf";
                });
            }

            bmsGlobalObj.$.ti_widget_vtabcontainer.addEventListener('vtab_click', function(event) {
                switch (bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex) {
                    case 0:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        break;
                    case 1:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        break;
                    case 2:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        bmsGlobalObj.fire("communicationControlPageSelected");
                        break;
                    case 3:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        bmsGlobalObj.fire("cellBalancingPageSelected");
                        break;
                    case 4:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        bmsGlobalObj.fire("protectorPageSelected");
                        break;
                    case 5:
                        bmsGlobalObj.fire("startFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        break;
                    case 6:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        bmsGlobalObj.fire("diagnosticsPageSelected");
                        break;
                    case 7:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        break;
                    case 8:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("startBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        break;
                    case 9:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        break;
                    case 10:
                        bmsGlobalObj.fire("stopFaultsRecursiveRead");
                        bmsGlobalObj.fire("stopBridgeControlRecursiveRead");
                        bmsGlobalObj.fire("closeGpioPopup");
                        break;
                }
            });

            document.addEventListener('routePage', function(e) {
                var page = e.detail.page;
                switch(page) {
                    case 'cellMonitor':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 1;
                        break;
                    case 'communicationControl':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 2;
                        break;
                    case 'cellBalancing':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 3;
                        break;
                    case 'protector':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 4;
                        break;
                    case 'faults':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 5;
                        break;
                    case 'diagnostics':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 6;
                        break;
                    case 'commandSequence':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 7;
                        break;
                    case 'bridgeControl':
                        bmsGlobalObj.$.ti_widget_vtabcontainer.selectedIndex = 8;
                }
            });


            // Expand vtabcontainer nav bar when user clicks on menu icon or 'Menu' label
            templateObj.toggleMenu = function(event){
                bmsGlobalObj.$.ti_widget_vtabcontainer.isExpanded = !bmsGlobalObj.$.ti_widget_vtabcontainer.isExpanded;
            };
            templateObj.$.ti_widget_icon_button_menu.addEventListener('click',templateObj.toggleMenu);
            templateObj.$.ti_widget_label_menu.addEventListener('click',templateObj.toggleMenu);

            var regPage = document.querySelector('ti-widget-register-page');
            if (regPage) {
                // get the active device name from the register page
                templateObj.$.ti_widget_menubar.productName = regPage.deviceName;

                // called when the user clicks the Write Register button:
                regPage.addEventListener("write_reg", function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                }.bind(this));
                
                // called when the user clicks the Write All Registers button:
                regPage.addEventListener("write_regs", function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                }.bind(this));                

                // called when the user clicks the Read Register button:
                regPage.addEventListener("read_reg", function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                }.bind(this));

                // called when the user clicks the Read All Registers button:
                regPage.addEventListener("read_regs", function (event) {
                    var detailsObj = event.detail;
                    if (detailsObj.type === "status") {
                        templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                    }
                }.bind(this));

                // called when the user changes the immediate / deferred mode
                regPage.addEventListener("update_mode_changed", function (event) {
                    var detailsObj = event.detail;
                    templateObj.$.ti_widget_eventlog_view.log("info", "request to change update mode to " + detailsObj.mode.label);
                }.bind(this));

                // called when the user changes the auto read refresh rate
                regPage.addEventListener("auto_read_config_changed", function (event) {
                    var detailsObj = event.detail;
                    templateObj.$.ti_widget_eventlog_view.log("info", "request to change automatic read interval to " + detailsObj.autoRead.label);
                }.bind(this));
            }

            var usbHidService = gc.services['ti-service-usbhid'];
            if (usbHidService) {
                usbHidService.addEventListener('usbHidError', function(event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify("error", detailsObj.info + ": " + detailsObj.error, false, true);
                })
            }

            templateObj.getRegisterData = function() {
                var regData = {
                    signature: "oneui-register-data",
                    data: []
                };
                var regPage = document.querySelector('ti-widget-register-page');
                var regBlocks = regPage.regBlocks;
                for (var i = 0; i < regBlocks.length; i++) {
                    var registersData = [];
                    for (var r = 0; r < regBlocks[i].registers.length; r++) {
                        var reg = regBlocks[i].registers[r];
                        var regName = reg.name.toLowerCase();
                        regName = regName.replace(/&/g, "_");
                        var item = {
                            idx: r,
                            id: regName,
                            value: reg.value
                        };
                        registersData.push(item);
                    }
                    regData.data.push(registersData);
                }
                var strJson = JSON.stringify(regData);
                return strJson;
            };

            templateObj.setRegisterData = function(strJson) {
                var regData = JSON.parse(strJson);
                var ok = true;
                var regPage = document.querySelector('ti-widget-register-page');
                var regBlocks = regPage._registerView.regBlocks;
                for (var b = 0; b < regBlocks.length; b++) {
                    for (var r = 0; r < regBlocks[b].registers.length; r++) {
                        var reg = regBlocks[b].registers[r];
                        var regName = reg.name.toLowerCase();
                        regName = regName.replace(/&/g, "_");
                        item = regData.data[b][r];
                        if (item.id === regName) {
                            reg.value = item.value;
                            if (!regPage.isDeferredWrite) {
                                reg.valueLastWritten = reg.value;
                                var intValue = gc.utils.getValueFromHexString("" + reg.value);
                                regPage.writeValueToTargetRegister(intValue, null, b, r);
                            } else {
                                regPage._registerView.refreshRegValue(b,r);
                            }
                        } else {
                            var msg = "load registers failed: register name mismatch (" + item.id + " does not match " + regName + ")";
                            templateObj.logMsgAndNotify("warning", msg, false, true);
                        }
                    }
                }
                if (regPage.isDeferredWrite){
                    regPage._registerView._deferredAccessCtr++; // update the page
                }
            };

            templateObj.showToastMsg = function(msgType, msg, durationInSeconds) {
                templateObj.$.ti_widget_toast.hideToast();
                templateObj.$.ti_widget_toast.message = msg;
                switch (msgType.toLowerCase()) {
                    case "error":
                        templateObj.$.ti_widget_toast.iconName = "error";
                        templateObj.$.ti_widget_toast.backgroundColor = "red";
                        break;
                    case "warning":
                        templateObj.$.ti_widget_toast.iconName = "warning";
                        templateObj.$.ti_widget_toast.backgroundColor = "orange";
                        break;
                    default:
                        templateObj.$.ti_widget_toast.iconName = "check";
                        templateObj.$.ti_widget_toast.backgroundColor = "teal";
                        break;
                }
                templateObj.$.ti_widget_toast.duration = +durationInSeconds;
                templateObj.$.ti_widget_toast.showToast();
                console.log("toast msg [" + msgType + "]: " + msg);
            };

            templateObj.logMsgAndNotify = function(msgType, msg, showInStatusBar, showInToast) {
                var iconName = templateObj.$.ti_widget_eventlog_view.iconForItem(msgType);
                if (showInStatusBar) {
                    var statusBar = gc.services['ti-widget-statusbar'];
                    if (statusBar) {
                        statusBar.statusString3 = msg;
                        statusBar.setIcon(iconName);
                    }
                }
                if (showInToast) {
                    templateObj.showToastMsg(msgType, msg, 4);
                }
                templateObj.$.ti_widget_eventlog_view.log(msgType, msg);
            };
            // templateObj.$.ti_widget_label_learnmore.addEventListener("click",function(event){
            //      templateObj.$.ti_widget_label_productname.fontColor="#444";
            //      templateObj.$.ti_widget_label_productdesc.fontColor="#444";
            //      templateObj.$.ti_widget_label_learnmore.setAttribute('hidden',true);
            //      templateObj.$.ti_widget_label_showless.removeAttribute('hidden');
            //      templateObj.$.ti_widget_vtab_home.style.backgroundColor = "white";
            //      templateObj.$.ti_widget_tabcontainer.selectedIndex = 1;
            // });
            // templateObj.$.ti_widget_label_showless.addEventListener("click",function(event){
            //      templateObj.$.ti_widget_label_productname.fontColor="white";
            //      templateObj.$.ti_widget_label_productdesc.fontColor="white";
            //      templateObj.$.ti_widget_label_learnmore.removeAttribute('hidden');
            //      templateObj.$.ti_widget_label_showless.setAttribute('hidden',true);
            //      templateObj.$.ti_widget_vtab_home.style.backgroundColor = "rgb(20, 140, 156)";
            //      templateObj.$.ti_widget_tabcontainer.selectedIndex = 0;
            // });
            if (templateObj.$.ti_widget_button_connect) {
                    templateObj.$.ti_widget_button_connect.addEventListener("click",function(event){
                        if (gc.connectionManager){
                            gc.connectionManager.connect();
                        }
                    });
            }


        }, 1);

    });
};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    document.addEventListener('DOMContentLoaded', init);
}